from unittest import TestCase
from .constants import get_persistent_dir
from tests.fixtures import init_dir, clear_dir


class GetPersistentDirTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_value(self):
        self.assertFalse((self.root / ".abstra" / "persistent").exists())
        self.assertEqual(self.root / ".abstra" / "persistent", get_persistent_dir())
        self.assertTrue((self.root / ".abstra" / "persistent").exists())



from __future__ import annotations  # required for type check
import datetime

from simple_websocket.ws import ConnectionClosed
import flask_sock, traceback, inspect
from typing import Literal, Optional, Dict, TYPE_CHECKING, Tuple
import traceback

from abstra_internals.compatibility import compat_traceback


from .execution import ABSTRA_RUN_KEY, Execution, RequestData, NoExecutionFound
from ..contract import forms_contract
from ..utils import deserialize, serialize
from ..utils.debug import make_debug_data
from ..contract import forms_contract
from ..jwt_auth import UserClaims
from ..repositories.execution_logs import FormEventLogEntry
from ..repositories.execution import ExecutionRepository
from ..repositories.execution_logs import ExecutionLogsRepository
from ..repositories.stage_run import StageRunRepository
from ..utils.environment import IS_PRODUCTION


if TYPE_CHECKING:
    from ..repositories.project.project import FormStage


class FormExecution(Execution):
    _connection: flask_sock.Server
    debug_enabled = True

    @classmethod
    def get_current_execution(cls) -> Optional["FormExecution"]:
        execution = Execution.get_current_execution()
        if isinstance(execution, FormExecution):
            return execution

        return None

    @staticmethod
    def broadcast(msg: forms_contract.Message):
        for e in list(Execution.executions.values()):
            if isinstance(e, FormExecution) and e.connected and e.closed == False:
                e.send(msg)

    def __init__(
        self,
        stage: "FormStage",
        is_initial: bool,
        connection: flask_sock.Server,
        request: RequestData,
        stage_run_repository: StageRunRepository,
        execution_repository: ExecutionRepository,
        execution_logs_repository: ExecutionLogsRepository,
        execution_id=None,
    ):
        self._connection = connection
        if execution_id is not None:
            self.id = execution_id

        super().__init__(
            stage,
            is_initial,
            request,
            stage_run_repository=stage_run_repository,
            execution_repository=execution_repository,
            execution_logs_repository=execution_logs_repository,
            execution_id=execution_id,
        )

    @property
    def connected(self) -> bool:
        return self._connection.connected

    def log_form_message(self, type: str, payload: dict):
        self.execution_logs_repository.save(
            FormEventLogEntry(
                execution_id=self.id,
                created_at=datetime.datetime.now(),
                event="form-message",
                payload={
                    "messageType": type,
                    "messagePayload": payload,
                },
                sequence=self.log_count,
            )
        )

    def send(self, msg: forms_contract.Message):
        self.log_form_message(msg.type, msg.data)

        skip_sending = isinstance(msg, forms_contract.StdioMessage) and IS_PRODUCTION
        if skip_sending:
            return

        if self.debug_enabled:
            if isinstance(msg, forms_contract.CloseMessage) and msg.close_dto.exception:
                debug = make_debug_data(msg.close_dto.exception)
            else:
                debug = make_debug_data(inspect.stack())
            msg.data = {**msg.data, **debug}

        str_data = serialize(msg.to_json())
        self._connection.send(str_data)

    def recv(self) -> Tuple[str, Dict]:
        try:
            str_data = self._connection.receive()
        except ConnectionClosed:
            self.close()
            return "close", {}

        data = deserialize(str_data)

        self.log_form_message(data["type"], data)
        return data["type"], data

    def close(self, reason: Optional[str] = ""):
        self._connection.close(message=reason)
        self.closed = True

    def stdio(self, event: Literal["stderr", "stdout"], text: str) -> None:
        self.send(forms_contract.StdioMessage(event, text))
        return super().stdio(event, text)

    def end(self):
        if self.connected:
            self.close(reason=traceback.format_exc() or None)

    def handle_lock_failed(self) -> None:
        status = self.stage_run.status if self.stage_run else None
        self.send(forms_contract.LockFailedMessage(status))

    # flows

    @property
    def query_params(self) -> Dict:
        return self.context.request.query_params

    def receive(self):
        while True:
            type, data = self.recv()
            if type in ["heartbeat", "browser:try-disconnect"]:
                continue

            return data

    def get_user(self, refresh: bool = False):
        # TODO: store/cache this
        self.send(forms_contract.AuthRequireInfoMessage(refresh=refresh))

        while True:
            type, data = self.recv()
            if type != "auth:saved-jwt":
                continue

            claims = UserClaims.from_jwt(data["jwt"])
            if not claims:
                self.send(forms_contract.AuthInvalidJWTMessage())
                continue

            self.send(forms_contract.AuthValidJWTMessage())
            return claims

    def execute_js(self, code: str, context: dict = {}):
        self.send(forms_contract.ExecuteJSRequestMessage(code, context))

        while True:
            type, data = self.recv()
            if type != "execute-js:response":
                continue

            return data.get("value")

    def alert(self, message: str, severity: str):
        severity = (
            severity if severity in ["info", "warn", "error", "success"] else "info"
        )
        self.send(forms_contract.AlertMessage(message, severity))

    def redirect(self, url: str, query_params: Optional[dict] = None):
        query_params = query_params if query_params is not None else self.query_params
        self.send(forms_contract.RedirectMessage(url, query_params))

    # forms

    def _recv_start(self):
        type = None
        data = None

        type, data = self.recv()
        if type != "start":
            return self._recv_start()

        return data["params"]

    def received_stage_run_id(self) -> Optional[str]:
        return self.context.request.query_params.get(ABSTRA_RUN_KEY)

    def handle_start(self) -> None:
        self.context.request.query_params = self._recv_start()
        self.send(forms_contract.ExecutionIdMessage(self.id))

    def handle_success(self) -> None:
        close_dto = forms_contract.CloseDTO(exit_status="SUCCESS")
        self.send(forms_contract.CloseMessage(close_dto))
        self.close()

    def handle_failure(self, e: Exception) -> None:
        self.send(
            forms_contract.StdioMessage(
                "stderr",
                "".join(
                    compat_traceback.format_exception(e),
                ),
            )
        )

        close_dto = forms_contract.CloseDTO(
            exit_status="EXCEPTION",
            exception=e,
        )
        self.send(forms_contract.CloseMessage(close_dto))

    def attempt_handle_exception(self, e: Exception) -> bool:
        NORMAL_CLOSURE = 1000
        GOING_AWAY = 1001

        if not isinstance(e, flask_sock.ConnectionClosed):
            return False

        if e.reason == NORMAL_CLOSURE:
            return True

        if e.reason == GOING_AWAY:
            self.status = "abandoned"
            return True

        self.log_form_message(
            "connection-closed",
            {
                "message": f"[ERROR] Connection closed with code {e.reason}: {e.message}\n",
                "reason": e.reason,
            },
        )

        return False


def get_form_execution_throwable() -> FormExecution:
    execution = FormExecution.get_current_execution()
    if not execution:
        raise NoExecutionFound()
    return execution



import flask

from ...usage import usage
from ..utils import send_from_dist

from ..controller.main import MainController
from ..controller import ai as ai_controller
from ..controller import jobs as jobs_controller
from ..controller import hooks as hooks_controller
from ..controller import forms as forms_controller
from ..controller import login as login_controller
from ..controller import kanban as kanban_controller
from ..controller import assets as assets_controller
from ..controller import scripts as scripts_controller
from ..controller import linters as linters_controller
from ..controller import debugger as debugger_controller
from ..controller import workspace as workspace_controller
from ..controller import workflows as workflows_controller
from ..controller import stage_runs as stage_runs_controller
from ..controller import requirements as requirements_controller
from ..controller import visualizations as visualizations_controller


def __get_api_bp(controller: MainController):
    bp = flask.Blueprint("editor_api", __name__)

    kanban_bp = kanban_controller.get_editor_bp()
    bp.register_blueprint(kanban_bp, url_prefix="/kanban")

    visualizations_bp = visualizations_controller.get_editor_bp()
    bp.register_blueprint(visualizations_bp, url_prefix="/visualizations")

    stage_run_bp = stage_runs_controller.get_editor_bp()
    bp.register_blueprint(stage_run_bp, url_prefix="/stage_runs")

    workspace_bp = workspace_controller.get_editor_bp(controller)
    bp.register_blueprint(workspace_bp, url_prefix="/workspace")

    forms_bp = forms_controller.get_editor_bp(controller)
    bp.register_blueprint(forms_bp, url_prefix="/forms")

    hooks_bp = hooks_controller.get_editor_bp(controller)
    bp.register_blueprint(hooks_bp, url_prefix="/hooks")

    jobs_bp = jobs_controller.get_editor_bp(controller)
    bp.register_blueprint(jobs_bp, url_prefix="/jobs")

    scripts_bp = scripts_controller.get_editor_bp(controller)
    bp.register_blueprint(scripts_bp, url_prefix="/scripts")

    workflows_bp = workflows_controller.get_editor_bp(controller)
    bp.register_blueprint(workflows_bp, url_prefix="/workflows")

    requirements_bp = requirements_controller.get_editor_bp(controller)
    bp.register_blueprint(requirements_bp, url_prefix="/requirements")

    debugger_bp = debugger_controller.get_editor_bp(controller)
    bp.register_blueprint(debugger_bp, url_prefix="/debugger")

    login_bp = login_controller.get_editor_bp(controller)
    bp.register_blueprint(login_bp, url_prefix="/login")

    linters_bp = linters_controller.get_editor_bp(controller)
    bp.register_blueprint(linters_bp, url_prefix="/linters")

    ai_bp = ai_controller.get_editor_bp(controller)
    bp.register_blueprint(ai_bp, url_prefix="/ai")

    assets_bp = assets_controller.get_editor_bp(controller)
    bp.register_blueprint(assets_bp, url_prefix="/assets")

    return bp


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor", __name__)

    api_bp = __get_api_bp(controller)
    bp.register_blueprint(api_bp, url_prefix="/api")

    @bp.get("/")
    @usage
    def _spa_index():
        return send_from_dist("editor.html", "editor.html")

    @bp.get("/<path:filename>")
    def _spa(filename: str):
        return send_from_dist(filename, "editor.html")

    return bp



from typing import Optional
from abstra_internals.constants import get_persistent_dir  # exported
from abstra_internals.execution.form_execution import get_form_execution_throwable


def get_user(refresh: bool = False):
    execution = get_form_execution_throwable()
    return execution.get_user(refresh=refresh)


def execute_js(code: str, context: dict = {}):
    execution = get_form_execution_throwable()
    return execution.execute_js(code, context=context)


def alert(message: str, severity: str = "info"):
    execution = get_form_execution_throwable()
    return execution.alert(message, severity)


def redirect(url: str, query_params: Optional[dict] = None):
    execution = get_form_execution_throwable()
    return execution.redirect(url, query_params)


def get_query_params() -> dict:
    execution = get_form_execution_throwable()
    return execution.query_params



from dataclasses import dataclass
from typing import Optional, Union, List, Literal, TypedDict, Dict


class ValidationResult(TypedDict):
    status: bool
    message: str


BrowserMessageTypes = Literal[
    "browser:try-disconnect",
    "auth:saved-jwt",
    "user-event",
    "form-response",
]

PythonMessageType = Literal[
    "stdout",
    "stderr",
    "auth:require-info",
    "auth:invalid-jwt",
    "auth:valid-jwt",
    "redirect",
    "alert",
    "execute-js:request",
    "files:changed",
    "stage-run:lock-failed",
    "execution-id",
    "program:end",
    "form-update",
    "form",
]

MessageType = Union[BrowserMessageTypes, PythonMessageType]


class Message:
    type: PythonMessageType
    data: dict

    def __init__(self, data: Dict):
        self.data = data

    def to_json(self) -> Dict[str, Union[str, dict]]:
        return {"type": self.type, **self.data}


class StdioMessage(Message):
    type: PythonMessageType

    def __init__(self, type: Literal["stdout", "stderr"], log: str):
        super().__init__({"log": log})
        self.type = type


class AuthRequireInfoMessage(Message):
    type = "auth:require-info"

    def __init__(self, refresh: bool):
        super().__init__({"refresh": refresh})


class AuthInvalidJWTMessage(Message):
    type = "auth:invalid-jwt"

    def __init__(self):
        super().__init__({})


class AuthValidJWTMessage(Message):
    type = "auth:valid-jwt"

    def __init__(self):
        super().__init__({})


class RedirectMessage(Message):
    type = "redirect"

    def __init__(self, url: str, query_params: dict):
        super().__init__({"url": url, "queryParams": query_params})


class AlertMessage(Message):
    type = "alert"

    def __init__(self, message: str, severity: str):
        super().__init__({"message": message, "severity": severity})


class ExecuteJSRequestMessage(Message):
    type = "execute-js:request"

    def __init__(self, code: str, context: dict):
        super().__init__({"code": code, "context": context})


class FilesChangedMessage(Message):
    type = "files:changed"

    def __init__(self):
        super().__init__({})


class LockFailedMessage(Message):
    type = "stage-run:lock-failed"

    def __init__(self, status: Optional[str]):
        super().__init__({"status": status})


@dataclass
class CloseDTO:
    exit_status: Literal["SUCCESS", "EXCEPTION"]
    exception: Optional[Union[Exception, None]] = None


class ExecutionIdMessage(Message):
    type = "execution-id"

    def __init__(self, execution_id: str):
        super().__init__({"executionId": execution_id})


class CloseMessage(Message):
    type = "program:end"
    close_dto: CloseDTO

    def __init__(self, close_dto: CloseDTO):
        self.close_dto = close_dto
        exc = str(close_dto.exception) if close_dto.exception else None
        super().__init__({"exitStatus": close_dto.exit_status, "exception": exc})


class FormUpdateMessage(Message):
    type = "form-update"

    def __init__(self, widgets: List, validation: ValidationResult, event_seq: int):
        super().__init__(
            {
                "widgets": widgets,
                "validation": validation,
                "seq": event_seq,
            }
        )


class FormMessage(Message):
    type = "form"

    def __init__(
        self,
        widgets: List,
        actions: List,
        end_program: bool,
        reactive_polling_interval: int,
        steps: List,
    ):
        super().__init__(
            {
                "widgets": widgets,
                "actions": actions,
                "endProgram": end_program,
                "reactivePollingInterval": reactive_polling_interval,
                "steps": steps,
            }
        )



import flask

from abstra_internals.repositories import (
    stage_run_repository,
    execution_repository,
    execution_logs_repository,
)
from ...usage import usage
from .main import MainController
from ..workflow_engine import workflow_engine
from ...execution.script_execution import ScriptExecution
from ..utils import is_it_true
from abstra_internals.repositories.execution_logs import (
    FormEventLogEntry,
)


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_scripts", __name__)

    @bp.get("/<path:id>")
    @usage
    def _get_script(id: str):
        script = controller.get_script(id)
        if not script:
            flask.abort(404)
        return script.editor_dto

    @bp.get("/")
    @usage
    def _get_scripts():
        return [f.editor_dto for f in controller.get_scripts()]

    @bp.post("/")
    @usage
    def _create_script():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        script = controller.create_script(title, file, workflow_position)
        return script.editor_dto

    @bp.put("/<path:id>")
    @usage
    def _update_script(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        script = controller.update_stage(id, data)
        return script.editor_dto if script else None

    @bp.delete("/<path:id>")
    @usage
    def _delete_script(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_script(id, remove_file)
        return {"success": True}

    @bp.post("/<path:id>/test")
    @usage
    def _test_script(id: str):
        script = controller.get_script(id)
        data = flask.request.json
        if not data:
            flask.abort(400)
        stage_run_id = data.get("stage_run_id")

        if not script or not stage_run_id:
            flask.abort(400)

        execution = ScriptExecution(
            stage=script,
            stage_run_id=stage_run_id,
            stage_run_repository=stage_run_repository,
            execution_repository=execution_repository,
            execution_logs_repository=execution_logs_repository,
        )

        execution.run()
        workflow_engine.notify_ran(execution)

        stdout: str = ""
        stderr: str = ""
        for entry in execution_logs_repository.get(execution.id):
            if isinstance(entry, FormEventLogEntry):
                continue

            text = entry.payload.get("text")
            if not text:
                continue

            if entry.event == "stdout":
                stdout += text
            elif entry.event == "stderr":
                stderr += text
            elif entry.event == "unhandled-exception":
                stderr += text

        return {"stdout": stdout, "stderr": stderr}

    return bp



from ....repositories.stage_run import StageRun
from ....repositories.project.project import IteratorStage


def iterator_strategy(
    stage: IteratorStage, stage_run: StageRun, transition_type="iterators:each"
):
    variable_value = stage_run.data.get(stage.variable_name)

    if not isinstance(variable_value, list):
        return []

    next_stage_ids = [
        t.target_id for t in stage.workflow_transitions if t.type == transition_type
    ]

    return [
        dict(stage=stage_id, data={"item": item})
        for item in variable_value
        for stage_id in next_stage_ids
    ]



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Union,
)
from abstra_internals.stdio_skip_request import safe_print

import requests
from pydantic.dataclasses import dataclass
from abstra_internals.utils.datetime import (
    from_utc_iso_string,
    to_utc_iso_string,
)

from ..utils.environment import SIDECAR_HEADERS, SIDECAR_URL
from ..utils.file import LOCAL_LOG_FOLDER


class LogEntryFactory:
    @staticmethod
    def from_dto(dto: Dict) -> "LogEntry":
        if dto.get("event") == "stdout" or dto.get("event") == "stderr":
            return StdioLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                event=dto["event"],
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        elif dto.get("event") == "form-message":
            return FormEventLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        elif dto.get("event") == "unhandled-exception":
            return UnhandledExceptionLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        else:
            raise Exception("Invalid log entry type")


@dataclass
class StdioLogEntry:
    execution_id: str
    created_at: datetime
    event: Literal["stderr", "stdout"]
    payload: Dict[Literal["text"], str]
    sequence: int

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


@dataclass
class FormEventLogEntry:
    execution_id: str
    created_at: datetime
    sequence: int
    payload: Dict[Literal["messageType", "messagePayload"], Union[str, Dict]]
    event: Literal["form-message"] = "form-message"

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


@dataclass
class UnhandledExceptionLogEntry:
    execution_id: str
    created_at: datetime
    payload: Dict[Literal["text"], str]
    sequence: int
    event: Literal["unhandled-exception"] = "unhandled-exception"

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


LogEntry = Union[StdioLogEntry, FormEventLogEntry, UnhandledExceptionLogEntry]
LogEvent = Literal["stdout", "stderr", "form-message", "unhandled-exception"]


class ExecutionLogsRepository(ABC):
    @abstractmethod
    def save(self, log_entry: LogEntry) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        raise NotImplementedError()


class LocalExecutionLogsRepository(ExecutionLogsRepository):
    _logs: Dict[str, List[LogEntry]]

    def __init__(self):
        self._logs = {}

    def save(self, log_entry: LogEntry) -> None:
        execution_id = log_entry.execution_id

        if not execution_id:
            return

        log_file = Path(LOCAL_LOG_FOLDER) / f"{execution_id}.log"

        if not log_file.parent.exists():
            log_file.parent.mkdir(parents=True)

        with open(log_file, "a", encoding="utf-8") as f:
            f.write(json.dumps(log_entry.to_dto()) + "\n")

    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        log_file = Path(LOCAL_LOG_FOLDER) / f"{execution_id}.log"

        if not log_file.exists():
            return []

        try:
            with open(log_file, "r", encoding="utf-8") as f:
                logs: List[LogEntry] = []
                for line in f.readlines():
                    dto = json.loads(line)
                    if event and dto["event"] != event:
                        continue
                    logs.append(LogEntryFactory.from_dto(dto))
                return logs
        except Exception as e:
            import traceback

            traceback.print_exc()
            print(f"Error reading log file {log_file}: {e}")
            return []


class RemoteExecutionLogsRepository(ExecutionLogsRepository):
    def __init__(self, url: str, headers: Dict[str, str]):
        self.url = url
        self.headers = headers

    def save(self, log_entry: LogEntry) -> None:
        res = requests.post(
            f"{self.url}/executions/{log_entry.execution_id}/logs",
            json=log_entry.to_dto(),
            headers=self.headers,
        )

        res.raise_for_status()

    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        response = requests.get(
            f"{self.url}/executions/{execution_id}/logs",
            params={"event": event} if event else None,
            headers=self.headers,
        )

        response.raise_for_status()

        return [LogEntryFactory.from_dto(log) for log in response.json()]


def execution_logs_repository_factory() -> ExecutionLogsRepository:
    if SIDECAR_URL:
        return RemoteExecutionLogsRepository(url=SIDECAR_URL, headers=SIDECAR_HEADERS)
    else:
        return LocalExecutionLogsRepository()
